/*;  khgrid.cc {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 2.1 $$ {\rhead{\sl khgrid.cc\quad\thepage}} */
/*;  $$Date: 1998/07/16 09:34:02 $$  */
/*;  $$Source: /home/shiiba/KhProgDevelop/KhC++Lib/matrix/khgrid.cc,v $$  */
/*;  $$RCShost: tone.kuciv.kyoto-u.ac.jp $$  */

#ifndef khgrid_cc
#define khgrid_cc

#include "khgrid.h"

template <class D>
KhGrid<D>::KhGrid()
{
  R = 1;
  C = 1;
  N = 1;

  Data = new D [N];

  if (Data == NULL)
    {
      kherror(); abort();
    }
}

template <class D>
KhGrid<D>::KhGrid
(
 size_t rows,
 size_t cols
 )
{
  R = rows;
  C = cols;
  N = rows * cols;

  if ((R == 0) || (C == 0))
    {
      kherror(); abort();
    }

  if ((N < R) || (N < C))
    {
      kherror(); abort();
    }

  Data = new D [N];

  if (Data == NULL)
    {
      kherror(); abort();
    }
}

template <class D>
KhGrid<D>::KhGrid
(
 const KhGrid<D> & m
 )
{
  R = m.R;
  C = m.C;
  N = m.N;

  Data = new D [N];

  if (Data == NULL)
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    Data[i] = m.Data[i];
}

// destructors
template <class D>
KhGrid<D>::~KhGrid()
{
  delete [] Data;
}

// assignment operator
template <class D>
void KhGrid<D>::operator =
(
 const KhGrid<D> & m
 )
{
  R = m.R;
  C = m.C;
  N = m.N;

  delete [] Data;

  Data = new D [N];

  if (Data == NULL)
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    Data[i] = m.Data[i];
}

template <class D>
void KhGrid<D>::operator =
(
 const D * array
 )
{
  // note: must assume that array has N members
  const D * aptr = array;
  D * mptr = Data;

  for (size_t i = 0; i < N; ++i)
    {
      (*mptr) = (*aptr);
      ++aptr;
      ++mptr;
    }
}

template <class D>
const D& KhGrid<D>::operator ()(
  size_t row,
  size_t col
  ) const
{
#ifndef NO_SUBCHK
  if (row > R || col > C)
    {
      cerr << "Invalid subscript. " << endl;
      kherror();  abort();
    }
#endif
  return Data[Index(row,col)];
}

template <class D>
D& KhGrid<D>::Ref(
  size_t row,
  size_t col
  )
{
#ifndef NO_SUBCHK
  if (row > R || col > C)
    {
      cerr << "Invalid subscript. " << endl;
      kherror();  abort();
    }
#endif
  return Data[Index(row,col)];
}

// apply a function to each element
template <class D>
void KhGrid<D>::Apply
(
 D (* func)(const D & n)
 )
{
  for (size_t i = 0; i < N; ++i)
    Data[i] = func(Data[i]);
}

template <class D>
KhGrid<D> Apply(
  const KhGrid<D> & m,
  D (* func)(const D & n)
  )
{
  KhGrid<D> result(m);

  for (size_t i = 0; i < m.N; ++i)
    result.Data[i] = func(result.Data[i]);

  return result;
}

// fill matrix with specific value
template <class D>
void KhGrid<D>::Fill(
  const D & x
  )
{
  for (size_t i = 0; i < N; ++i)
    Data[i] = x;
}

// create a row matrix
template <class D>
KhGrid<D> KhGrid<D>::VectorRow(
  size_t row
  )
{
  if (row >= R)
    {
      kherror(); abort();
    }

  KhGrid<D> vector(C,1);

  for (size_t i = 0; i < C; ++i)
    vector.Ref(i,0) = Data[Index(row,i)];

  return vector;
}

// create a column vector
template <class D>
KhGrid<D> KhGrid<D>::VectorCol(
  size_t col
  )
{
  if (col >= C)
    {
      kherror(); abort();
    }

  KhGrid<D> vector(1,R);

  for (size_t i = 0; i < R; ++i)
    vector.Ref(0,i) = Data[Index(i,col)];

  return vector;
}

// transpose a matrix
template <class D>
KhGrid<D> KhGrid<D>::Transpose()
{
  KhGrid<D> result(C,R);

  const D * tptr = Data;
  D * rptr;

  for (size_t i = 0; i < R; ++i)
    {
      rptr = result.Data + i;

      for (size_t j = 0; j < C; ++j)
	{
	  (*rptr) = (*tptr);
	  ++tptr;
	  rptr += R;
	}
    }

  return result;
}

template <class D>
void KhGrid<D>::Resize(
  size_t rows,
  size_t cols
  )
{
  delete [] Data;

  R = rows;
  C = cols;
  N = rows * cols;

  if ((R == 0) || (C == 0))
    {
      kherror(); abort();
    }

  if ((N < R) || (N < C))
    {
      kherror(); abort();
    }

  Data = new D [N];

  if (Data == NULL)
    {
      kherror(); abort();
    }
}

#endif

/*; end of file */
