/*;  khmatrix.cc {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.3 $$ {\rhead{\sl khmatrix.cc\quad\thepage}} */
/*;  $$Date: 1998/09/26 11:09:31 $$  */
/*;  $$Source: /home/shiiba/KhProgDevelop/KhC++Lib/matrix/khmatrix.cc,v $$  */
/*;  $$RCShost: tone.kuciv.kyoto-u.ac.jp $$  */

#ifndef khmatrix_cc
#define khmatrix_cc

#include "khsimple.h"
#include "khmatrix.h"
#include "khetype.h"

template <class D>
KhMatrix<D>::KhMatrix(
  size_t rows,
  size_t cols,
  const D & init
  )
  : KhGrid<D>(rows,cols)
{
  for (size_t i = 0; i < N; ++i)
    Data[i] = init;
}

// interrogation
template <class D>
KhBoolean KhMatrix<D>::IsZero() const
{
  const D * ptr = Data;

  for (size_t i = 0; i < N; ++i)
    {
      if ((*ptr) != (D)0)
	return NO;

      ++ptr;
    }

  return YES;
}

template <class D>
KhBoolean KhMatrix<D>::IsDiagonal() const
{
  if (C != R)
    return NO;

  const D * ptr = Data;

  for (size_t ir = 0; ir < R; ++ir)
    {
      for (size_t ic = 0; ic < C; ++ic)
	{
	  if (ir == ic)
	    {
	      if ((*ptr) == (D)0)
		return NO;
	    }
	  else
	    {
	      if ((*ptr) != (D)0)
		return NO;
	    }

	  ++ptr;
	}
    }

  return YES;
}

template <class D>
KhBoolean KhMatrix<D>::IsIdentity() const
{
  if (C != R)
    return NO;

  const D * ptr = Data;

  for (size_t ir = 0; ir < R; ++ir)
    {
      for (size_t ic = 0; ic < C; ++ic)
	{
	  if (ir == ic)
	    {
	      if ((*ptr) != (D)1)
		return NO;
	    }
	  else
	    {
	      if ((*ptr) != (D)0)
		return NO;
	    }

	  ++ptr;
	}
    }

  return YES;
}

template <class D>
KhBoolean KhMatrix<D>::IsTridiagonal() const
{
  if ((C != R) || (C < 3))
    return NO;

  const D * ptr = Data;

  for (size_t ir = 0; ir < R; ++ir)
    {
      for (size_t ic = 0; ic < C; ++ic)
	{
	  if (ir != ic)
	    {
	      if (ir > ic)
		{
		  if ((ir - ic) > 1)
		    {
		      if ((*ptr) != (D)0)
			return NO;
		    }
		}
	      else
		{
		  if ((ic - ir) > 1)
		    {
		      if ((*ptr) != (D)0)
			return NO;
		    }
		}
	    }

	  ++ptr;
	}
    }

  return YES;
}

template <class D>
KhBoolean KhMatrix<D>::IsUpperTriangular() const
{
  if ((C != R) || (C < 2))
    return NO;

  size_t steps = 1;
  const D * ptr = Data + C;

  for (size_t ir = 1; ir < R; ++ir)
    {
      for (size_t s = 0; s < steps; ++s)
	{
	  if (ptr[s] != (D)0)
	    return NO;
	}

      ++steps;
      ptr += C;
    }

  return YES;
}

template <class D>
KhBoolean KhMatrix<D>::IsLowerTriangular() const
{
  if ((C != R) || (C < 2))
    return NO;

  size_t steps = C - 1;
  const D * ptr = Data;

  for (size_t ir = 1; ir < R; ++ir)
    {
      for (size_t s = steps; s > 0; --s)
	{
	  if (ptr[s] != (D)0)
	    return NO;
	}

      --steps;
      ptr += C + 1;
    }

  return YES;
}

template <class D>
KhBoolean KhMatrix<D>::IsPermutation() const
{
  if (C != R)
    return NO;

  char * ctags = new char[C];

  if (ctags == NULL)
    {
      kherror(); abort();
    }

  char * rtags = new char[R];

  if (rtags == NULL)
    {
      kherror(); abort();
    }

  memset(ctags,0,C);
  memset(rtags,0,R);

  KhBoolean result = YES;

  const D * ptr = Data;

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  if ((*ptr))
	    {
	      if (((*ptr) > 1)
		  ||  (rtags[ri] == 1)
		  ||  (ctags[ci] == 1))
		{
		  result = NO;
		  goto finished;
		}

	      rtags[ri] = 1;
	      ctags[ci] = 1;
	    }

	  ++ptr;
	}
    }

  // a goto label!
 finished:

  delete [] ctags;
  delete [] rtags;
  return result;
}

// addition
template <class D>
KhMatrix<D> KhMatrix<D>::operator +
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<D> result(m);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] += Data[i];

  return result;
}

template <class D>
void KhMatrix<D>::operator +=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    Data[i] += m.Data[i];
}

template <class D>
KhMatrix<D> KhMatrix<D>::operator +
(
 const D & x
 )
{
  KhMatrix<D> result(*this);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] += x;

  return result;
}

template <class D>
void KhMatrix<D>::operator +=
(
 const D & x
 )
{
  for (size_t i = 0; i < N; ++i)
    Data[i] += x;
}

// subtraction
template <class D>
KhMatrix<D> KhMatrix<D>::operator -
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<D> result(*this);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] -= m.Data[i];

  return result;
}

template <class D>
void KhMatrix<D>::operator -=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    Data[i] -= m.Data[i];
}

template <class D>
KhMatrix<D> KhMatrix<D>::operator -
(
 const D & x
 )
{
  KhMatrix<D> result(*this);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] -= x;

  return result;
}

template <class D>
void KhMatrix<D>::operator -=
(
 const D & x
 )
{
  for (size_t i = 0; i < N; ++i)
    Data[i] -= x;
}

// scalar product
template <class D>
KhMatrix<D> KhMatrix<D>::operator *
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<D> result(m);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] *= Data[i];

  return result;
}

template <class D>
void KhMatrix<D>::operator *=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    Data[i] *= m.Data[i];
}

template <class D>
KhMatrix<D> KhMatrix<D>::operator *
(
 const D & x
 )
{
  KhMatrix<D> result(*this);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] *= x;

  return result;
}

template <class D>
void KhMatrix<D>::operator *=
(
 const D & x
 )
{
  for (size_t i = 0; i < N; ++i)
    Data[i] *= x;
}

// division
template <class D>
KhMatrix<D> KhMatrix<D>::operator /
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<D> result(*this);

  for (size_t i = 0; i < N; ++i)
    {
      if (m.Data[i] == (D)0)
	{
	  kherror(); abort();
	}

      result.Data[i] /= m.Data[i];
    }

  return result;
}

template <class D>
void KhMatrix<D>::operator /=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    {
      if (m.Data[i] == (D)0)
	{
	  kherror(); abort();
	}

      Data[i] /= m.Data[i];
    }
}

template <class D>
KhMatrix<D> KhMatrix<D>::operator /
(
 const D & x
 )
{
  if (x == (D)0)
    {
      kherror(); abort();
    }

  KhMatrix<D> result(*this);

  for (size_t i = 0; i < N; ++i)
    result.Data[i] /= x;

  return result;
}

template <class D>
void KhMatrix<D>::operator /=
(
 const D & x
 )
{
  if (x == (D)0)
    {
      kherror(); abort();
    }

  for (size_t i = 0; i < N; ++i)
    Data[i] /= x;
}

// matrix multiplication
template <class D>
KhMatrix<D> KhMatrix<D>::operator %
(
 const KhMatrix<D> & m
 )
{
  if (C != m.R)
    {
      kherror(); abort();
    }

  KhMatrix<D> result(R,m.C);

  D * rptr = result.Data;
  const D * tptr, * mptr;

  for (size_t i = 0; i < R; ++i)
    {
      for (size_t j = 0; j < m.C; ++j)
	{
	  tptr =   Data + (i * R);
	  mptr = m.Data +  j;

          // ɤ(Cˤ륢르ꥺŵ, p.125)
          // 򤷤Ƥޤ
	  D sum = D(0);
/*
          D remainder = (D)0, t = (D)0;

	  for (size_t k = 0; k < C; ++k)
	    {
	      remainder += (*tptr) * (*mptr);
	      t = sum;
	      sum += remainder;
	      t -= sum;
	      remainder += t;

	      ++tptr;
	      mptr += m.C;
	    }
*/
	  for (size_t k = 0; k < C; k++) {
	    sum += Data[Index(i,k)] * m(k,j);
	  }
	  (*rptr) = sum;
	  ++rptr;
	}
    }

  return result;
}

// comparison operators
template <class D>
KhBoolean KhMatrix<D>::Equals
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  if (Data[Index(ri,ci)] != m.Data[Index(ri,ci)])
	    return NO;
	}
    }

  return YES;
}

template <class D>
KhMatrix<int> KhMatrix<D>::operator ==
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<int> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  result.Ref(ri,ci) =
	    (Data[Index(ri,ci)] == m.Data[Index(ri,ci)]);
	}
    }

  return result;
}

template <class D>
KhMatrix<int> KhMatrix<D>::operator !=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<int> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  result.Ref(ri,ci) =
	    (Data[Index(ri,ci)] != m.Data[Index(ri,ci)]);
	}
    }

  return result;
}

template <class D>
KhMatrix<int> KhMatrix<D>::operator <
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<int> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  result.Ref(ri,ci) =
	    (Data[Index(ri,ci)] < m.Data[Index(ri,ci)]);
	}
    }

  return result;
}

template <class D>
KhMatrix<int> KhMatrix<D>::operator <=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<int> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  result.Ref(ri,ci) =
	    (Data[Index(ri,ci)] <= m.Data[Index(ri,ci)]);
	}
    }

  return result;
}

template <class D>
KhMatrix<int> KhMatrix<D>::operator >
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<int> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  result.Ref(ri,ci) =
	    (Data[Index(ri,ci)] > m.Data[Index(ri,ci)]);
	}
    }

  return result;
}

template <class D>
KhMatrix<int> KhMatrix<D>::operator >=
(
 const KhMatrix<D> & m
 )
{
  if ((R != m.R) || (C != m.C))
    {
      kherror(); abort();
    }

  KhMatrix<int> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	{
	  result.Ref(ri,ci) =
	    (Data[Index(ri,ci)] >= m.Data[Index(ri,ci)]);
	}
    }

  return result;
}

// negate a matrix
template <class D>
KhMatrix<D> KhMatrix<D>::operator - ()
{
  KhMatrix<D> result(R,C);

  for (size_t ri = 0; ri < R; ++ri)
    {
      for (size_t ci = 0; ci < C; ++ci)
	result.Ref(ri,ci) = -(Elem(ri,ci));
    }

  return result;
}

// change size (destroying contents)
template <class D>
void KhMatrix<D>::Resize
(
 size_t rows,
 size_t cols,
 const D & init
 )
{
  this->KhGrid<D>::Resize(rows,cols);

  for (size_t i = 0; i < N; ++i)
    Data[i] = init;
}

// inner products
template <class D>
D KhMatrix<D>::InnerProduct
(
 const KhMatrix<D> & m
 ) const
{
  D result = (D)0;

  if (((R == 1) && (m.C == 1) && (C == m.R))
      ||  ((C == 1) && (m.R == 1) && (R == m.C)))
    {
      const D * ptr1 = Data;
      const D * ptr2 = m.Data;

      size_t end = (C == 1 ? R : C);

      // ɤ(Cˤ륢르ꥺŵ, p.125)
      // 򤷤Ƥޤ
      D remainder = (D)0, sum = (D)0, t = (D)0;
      for (size_t n = 0; n < end; ++n)
	{
	  remainder += (*ptr1) * (*ptr2);
	  t = sum;
	  sum += remainder;
	  t -= sum;
	  remainder += t;
	  ++ptr1;
	  ++ptr2;
	}
      result = sum;
    }
  else
    {
      kherror(); abort();
    }

  return result;
}

// calculate euclidean norm
// euclidean norm
template <class D>
double KhMatrix<D>::Norm() const
{
  const D* ptr = Data;
  size_t end = R*C;
  D remainder = (D)0, sum = (D)0, t = (D)0;
  for (size_t n = 0; n < end; ++n) {
    remainder += (*ptr) * (*ptr);
    t = sum;
    sum += remainder;
    t -= sum;
    remainder += t;
    ++ptr;
  }
  double result = sum;
  return sqrt(result);
}

// calculate determinant value
template <class D>
D KhMatrix<D>::Determinant() const
{
  if (R != C)
    {
      kherror(); abort();
    }

  if (R == 1)
    return (*Data);

  return DetRecursive();
}

template <class D>
D KhMatrix<D>::DetRecursive() const
{
  if (R == 2)
    {
      return   (    (*Data)    * (*(Data + 3)) )
	- ( (*(Data + 1)) * (*(Data + 2)) );
    }

  D result      = (D)0;
  const D * ptr = Data;

  for (size_t x = 0; x < R; ++x)
    {
      if (x & 1) // if on an even column
	result -= (*ptr) * (Minor(0,x)).DetRecursive();
      else
	result += (*ptr) * (Minor(0,x)).DetRecursive();

      ++ptr;
    }

  return result;
}

// create a minor matrix
template <class D>
KhMatrix<D> KhMatrix<D>::Minor
(
 size_t rdel,
 size_t cdel
 ) const
{
  if ((R != C) || (R < 2))
    {
      kherror(); abort();
    }

  KhMatrix<D> result(R-1,C-1);

  const D * psrc = Data;
  D * pdest = result.Data;

  for (size_t rsrc = 0; rsrc < R; ++rsrc)
    {
      if (rsrc != rdel)
	{
	  for (size_t csrc = 0; csrc < C; ++csrc)
	    {
	      if (csrc != cdel)
		{
		  *pdest = *psrc;
		  ++pdest;
		}

	      ++psrc;
	    }
	}
      else
	psrc += C;
    }

  return result;
}

// solve a system of linear equations via Gaussian elimination

template <class D>
KhBoolean KhMatrix<D>::LinSolve(KhMatrix<D>& X)
{
  if ((C - R) != 1)
    {
      cerr << "The number of columns is not appropriate." << endl;
      kherror(); abort();
    }

  size_t i, j, k, max;
  D temp;

  // forward elimination
  for (i = 0; i < R; ++i)
    {
      max = i;

      for (j = i + 1; j < R; ++j)
	{
	  if (KhAbsVal(Elem(j,i)) > KhAbsVal(Elem(max,i)))
	    max = j;
	}

      for (k = i; k < C; ++k)
	{
	  temp = Elem(i,k);
	  Elem(i,k) = Elem(max,k);
	  Elem(max,k) = temp;
	}

      if (Elem(i,i) == 0) return NO; // the matrix is singular

      for (j = i + 1; j < R; ++j)
	{
	  for (k = R; k >= i; --k)
	    {
	      Elem(j,k) -=   Elem(i,k)
		* Elem(j,i)
		/ Elem(i,i);

	      if (k == 0)
		break;
	    }
	}
    }

  // backward substitution
  for (j = R - 1; ; --j)
    {
      temp = (D)0;

      for (k = j + 1; k < R; ++k)
	temp += Elem(j,k) * X.Elem(k,0);

      X.Elem(j,0) = (Elem(j,R) - temp) / Elem(j,j);

      if (j == 0)
	break;
    }

  return YES;
}

template <class D>
ostream& operator<<(ostream& s, KhMatrix<D>& a)
{
  int i, j;   int m = (int)a.GetRows();   int n = (int)a.GetCols();
  const int ncol = 5;	// number of columes to be printed in one line

  int n1, n2 = 0;
  while (n2 < n) {
    n1 = n2;
    n2 = n1 + ncol;
    if (n2 > n) n2 = n;

    s << "      ";
    for (j = n1; j < n2; ++j)
      s << setw(10) << j + 1 << ' ';

    s << '\n';
    s << "------";
    for (j = n1; j < n2; ++j)
      s << "-----------";

    s << '\n';
    for (i = 0; i < m; ++i) {
      s << setw(4) << i + 1 << ": ";

      for (j = n1; j < n2; ++j) {
	s << setw(10) << setprecision(4) << a.Elem(i,j) << ' ';
      }
      s << '\n';
    }
  }
  return s;
}

template <class D>
istream& operator>>(istream& s, KhMatrix<D>& a)
{
  size_t nrow = a.GetRows(),   ncol = a.GetCols();

  char* type[] = {"integer", "long", "float", "double"};
  cout.flush();
  int ntype = ElementType((D)0);
  cout << "Input a " << type[ntype] << " matrix with the size "
       << nrow << " x " << ncol << " :" << endl;

  for (size_t i = 1; i <= nrow; i++) {
    cout << i ;
    if (i == 11 || i == 12) {
      cout << "th row: ";
    }
    else {
      size_t k = i % 10;
      size_t j = i % 100;
      if (k == 1 && j != 11)
      {
	cout << "st row: ";
      }
      else if (k == 2 && j != 12)
      {
	cout << "nd row: ";
      }
      else if (k == 3 && j != 13)
      {
	cout << "rd row: ";
      }
      else
      {
	cout << "th row: ";
      }
    }
    cout.flush();
    for (size_t j = 0; j < ncol; j++) {
      cin >> a.Elem(i-1, j);
    }
    cin.ignore(INT_MAX, '\n');
  }
  return s;
}

// =====================================================================

template <class Type>
Lufact<Type>::Lufact(const KhMatrix<Type>& a)
{
  int m = a.GetRows();
  int n = a.GetCols();
  if (m <= 1 || m != n)
    {
      kherror(); abort();
    }

  Alu.Resize(n, n);
  Alu = a;					// copy values from a
  Indx.Resize(n, 1);

  int i,imax,j,k;
  Type big,dum,sum,temp;

  KhMatrix<Type> vv(n, 1);
  D = (Type)1;
  for (i = 0; i < n; i++) {
    big = (Type)0;
    for (j = 0; j < n; j++)
      if ((temp = KhAbsVal(Alu.Elem(i,j))) > big) big = temp;
    if (big == (Type)0) {
      D = (Type)0;   return;	// d = 0 means a is singular
    }
    vv.Ref(i, 0) = (Type)1/big;
  }
  for (j = 0; j < n; j++) {
    for (i = 0; i < j; i++) {
      sum = Alu.Elem(i, j);
      for (k = 0; k < i; k++) sum -= Alu.Elem(i, k) * Alu.Elem(k, j);
      Alu.Elem(i, j) = sum;
    }
    big = (Type)0;
    for (i = j; i < n; i++) {
      sum = Alu.Elem(i, j);
      for (k = 0; k < j; k++)
	sum -= Alu.Elem(i,k) * Alu.Elem(k,j);
      Alu.Elem(i, j) = sum;
      if ( (dum = vv(i, 0)*KhAbsVal(sum)) >= big) {
	big = dum;
	imax = i;
      }
    }
    if (j != imax) {
      for (k = 0; k < n; k++) {
	dum = Alu.Elem(imax, k);
	Alu.Elem(imax, k) = Alu.Elem(j, k);
	Alu.Elem(j, k) = dum;
      }
      D  = -D;
      vv.Ref(imax, 0) = vv(j, 0);
    }
    Indx.Ref(j, 0) = imax;
    if (Alu.Elem(j, j) == (Type)0) {
      D = 0; return;
    }
    if (j != n) {
      dum = (Type)1/(Alu.Elem(j, j));
      for (i = j+1; i < n; i++) Alu.Elem(i, j) *= dum;
    }
  }
//  cout << Alu;
//  cout << Indx;
}

template <class Type>
KhBoolean Lufact<Type>::IsRegular()
{
  if (D == (Type)0)
    return NO;
  else
    return YES;
}

template <class Type>
KhBoolean Lufact<Type>::IsSingular()
{
  if (D == (Type)0)
    return YES;
  else
    return NO;
}

template <class Type>
Type Lufact<Type>::Determinant()
{
  if (D == (Type)0)
    return (Type)0;
  else {
    Type det = D;
    int n = Alu.GetRows();
    for (int i = 0; i < n; i++) {
      det *= Alu.Elem(i,i);
    }
    return det;
  }
}

template <class Type>
void Lufact<Type>::Solve(KhMatrix<Type>& b)
{
  int i,ip,j;
  int n = Alu.GetRows();
  Type sum;

  for (i = 0; i < n ;i++) {
    ip = Indx(i, 0);
    sum = b(ip, 0);
    b.Ref(ip, 0) = b(i, 0);
    for (j = 0; j < i; j++) sum -= Alu.Elem(i,j)*b(j, 0);
    b.Ref(i, 0)=sum;
  }

  for (i = n-1; i >= 0; i--) {
    sum=b(i, 0);
    for (j = i+1; j < n;j++) sum -= Alu.Elem(i,j)*b(j, 0);
    b.Ref(i, 0) = sum/Alu.Elem(i,i);
  }
}

template <class Type>
KhBoolean Lufact<Type>::Invmatrix(KhMatrix<Type>& inva)
{
  int n = Alu.GetRows();			// Կ
  KhMatrix<Type> col(n, 1);

  if (D == (Type)0) return NO;

  for (int j = 0; j < n; j++) {
    int i;
    for (i = 0; i < n; i++) col.Elem(i, 0) = (Type)0;
    col.Ref(j, 0) = (Type)1;
    Solve(col);
    for (i = 0; i < n; i++) inva.Elem(i,j) = col.Elem(i, 0);
  }
  return YES;
}

#endif

/*; end of file */
